/*	immanager - Text Input Method Manager
 *
 *  gtkimcontextptim.c - 
 *
 *	Authors: YE Nan <nan.ye@orange-ftgourp.com>,
 *           LUO Lei <Srella@163.com>
 *
 *	
 *	This software and associated documentation files (the "Software") 
 *	are copyright (C) 2005 LiPS Linux Phone Standards Forum [FranceTelecom] 
 *	All Rights Reserved. 
 *
 *	A copyright license is hereby granted for redistribution and use of 
 *	the Software in source and binary forms, with or without modification, 
 *	provided that the following conditions are met: 
 *	- Redistributions of source code must retain the above copyright notice, 
 *	this copyright license and the following disclaimer. 
 *  - Redistributions in binary form must reproduce the above copyright 
 * 	notice, this copyright license and the following disclaimer in the 
 *	documentation and/or other materials provided with the distribution. 
 *	- Neither the name of LiPS  nor the names of its Members may be used 
 *	to endorse or promote products derived from the Software without 
 *	specific prior written permission. 
 *
 *	A patent license for any Necessary Claims owned by Members of LiPS Forum 
 *	to make, have made, use, import, offer to sell, lease and sell or otherwise 
 *	distribute any implementation compliant with the any specification adopted 
 *	by the LiPS Forumcan be obtained from the respective Members on reasonable 
 *	and non-discriminatory terms and conditions and under reciprocity, as 
 *	regulated in more detail in the Internal Policy of the LiPS Forum. 
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER, ITS MEMBERS AND CONTRIBUTORS 
 *	"AS IS", AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *	THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE 
 *	AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER, 
 *	ITS MEMBERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 *	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 *	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *	POSSIBILITY OF SUCH DAMAGE. 
 */

#include <stdio.h>
#include <stdlib.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <iac.h>

#include "gtkimcontextptim.h"
#include "imptim.h"

#define PTIM_DBUS_OBJECT			"/org/gpe/status/ptim"
#define PTIM_DBUS_INTERFACE		"org.gpe.status.ptim"
#define PTIM_DBUS_METHOD 			"ptimindicator"

#define LONG_KEY_PRESS_INTERVAL 700

gboolean longkeyhandled = FALSE;
guint longkeypresstimer;			/*id of timer*/
int lastkeyval;
int candidatesnumberEN		= 8;		/*the restriction of the number of candidates displayed on the screen at the same time*/
int characternumberEN		= 0;		/*the length of the candidates currently*/
GKeyFile * keyfile;
static gint im_switch_key  = GDK_comma;

extern	GtkIMContextPTIM *	global_ic_ptim;

extern	const IMM_Functions gtk_imm_functions;

extern gboolean	(*engine_register)	(IMM_Functions * imm, IME_Functions	 * ime);
extern gboolean	(*helper_register)	(IMM_Functions * imm, IMH_Functions	 * imh);


static gboolean 
gtk_im_context_ptim_is_num_input (GtkIMContext * this)
{
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);
	IME_InputMethod * im_info;
	gboolean bIsNumInput;
	
#ifdef	DEBUG
	assert(this_ptim->ptim_engines_array);
#endif
	
	im_info = &g_array_index(this_ptim->ptim_engines_array, IME_InputMethod, this_ptim->ptim_engines_index);
	bIsNumInput = (gboolean) (im_info->capability_mask & CAP_NUM_INPUT);
	g_printf("IsNumInput:%d\n",bIsNumInput);
	return bIsNumInput;
}

static gboolean 
gtk_im_context_ptim_is_TS_eng_input (GtkIMContext * this)
{
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);
	IME_InputMethod * im_info;
	gboolean bIsTSengInput;

#ifdef	DEBUG
	assert(this_ptim->ptim_engines_array);
#endif

	im_info = &g_array_index(this_ptim->ptim_engines_array, IME_InputMethod, this_ptim->ptim_engines_index);
	bIsTSengInput = (gboolean) (im_info->capability_mask & CAP_TS_INPUT);
	g_printf("IsTSengInput:%d\n",bIsTSengInput);
	return bIsTSengInput;
}

static gboolean 
gtk_im_context_ptim_is_fcitx_input (GtkIMContext * this)
{
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);
	IME_InputMethod * im_info;
	gboolean bIsFcitxInput;

#ifdef	DEBUG
	assert(this_ptim->ptim_engines_array);
#endif

	im_info = &g_array_index(this_ptim->ptim_engines_array, IME_InputMethod, this_ptim->ptim_engines_index);
	bIsFcitxInput = (gboolean) (im_info->capability_mask & CAP_TS_FCITX_INPUT);
	g_printf("IsFcitxInput:%d\n",bIsFcitxInput);
	return bIsFcitxInput;
}


static gboolean 
gtk_im_context_ptim_switch_input_method(GtkIMContext * this)
{
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);
	IME_InputMethod * im_info;
	GString * icon_path;
	int helpercount;
	int i;

	g_print("%s(): entering\n", __FUNCTION__);

	global_ic_ptim = this_ptim;
	candidatesnumberEN = 8;
	characternumberEN = 0;	

	if (this_ptim->input_context != NULL)
	{
		this_ptim->engine_module->engine_func.destroy_input_context(this_ptim->input_context);

		g_free(this_ptim->input_context->uuid);
		g_free(this_ptim->input_context->charset);
		g_free(this_ptim->input_context);
	}

	this_ptim->input_context = g_new0(IME_InputContext, 1);

	im_info = &g_array_index(this_ptim->ptim_engines_array, IME_InputMethod, this_ptim->ptim_engines_index);
	
	this_ptim->input_context->uuid = g_strdup(im_info->uuid);
	this_ptim->input_context->charset = g_strdup(im_info->encoding);
	
	this_ptim->engine_module->engine_func.create_input_context(this_ptim->input_context);

	if (TRUE)				/*switch icon*/
	{
		iac_context_t * context = im_ptim_get_iac_contact();
		iac_arguments_t * args = NULL;
				
		args = iac_arguments_new();
		
		if (strstr(im_info->languages, "en_US") || strstr(im_info->languages, "en_UK"))
		{
			iac_arguments_append_args(args,
																IAC_TYPE_INT32,
																1,
																IAC_TYPE_INVALID);	
		}
		else if (g_strcasecmp(im_info->languages, "zh_CN") == 0)
		{
			iac_arguments_append_args(args,
																IAC_TYPE_INT32,
																2,
																IAC_TYPE_INVALID);	
		}
		else
		{
			iac_arguments_append_args(args,
																IAC_TYPE_INT32,
																0,
																IAC_TYPE_INVALID);	
		}
		
		iac_emit_signal(context,
										PTIM_DBUS_OBJECT,
										PTIM_DBUS_INTERFACE,
										PTIM_DBUS_METHOD,
										args);
										
		iac_arguments_free(args);
	}

	if(gtk_im_context_ptim_is_num_input(this))			/*number input			index=0 */
	{
		gtk_imm_functions.hide_all_candidates_list(this_ptim->input_context);
		return;
	}

	gchar *correspondinghelper;
	correspondinghelper = g_key_file_get_value(keyfile, "IM Group", 
											this_ptim->engine_list[this_ptim->ptim_engines_index], NULL);

	this_ptim->helper_module->helper_func.initialize(correspondinghelper);

  if(gtk_im_context_ptim_is_TS_eng_input(this) || gtk_im_context_ptim_is_fcitx_input(this))
  {
		this_ptim->status = IME_NORMAL;
		this_ptim->helper_module->helper_func.notify_ime_status(this_ptim->status);
		gtk_imm_functions.show_candidates_list(this_ptim->input_context);	
	}
	free(correspondinghelper);

	return;
}

static void 
gtk_im_context_ptim_init(GTypeInstance * this, gpointer class)
{
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);
	IME_InputMethod * im_info;
	IME_Helper_Info * imh_info;
	gint i;
	gchar *str;
	
	/* export Gloabl IMM Functions */
	global_ic_ptim = this_ptim; 
	
	g_print("%s(): entering\n", __FUNCTION__);

	/* GtkIMContextPTIM specific fields initialization. */
	this_ptim->client_window							= NULL;
	this_ptim->input_context							= NULL;
	this_ptim->preedit_candidates_list	= NULL;
	this_ptim->candidates_list						= NULL;
	this_ptim->status												= IME_PREEDIT;
	this_ptim->numeric_select							= FALSE;
	
	keyfile = g_key_file_new();
	if(!g_key_file_load_from_file(keyfile, getenv("PTIM_CONFIG_FILE"), 0, NULL))
	{
		g_print("can't open im config file!\n");
	}

	this_ptim->engine_list = g_key_file_get_keys(keyfile, "IM Group", &(this_ptim->ptim_engines_count), NULL);

	int j = 0;
	for (i = 0; this_ptim->engine_list[i] != NULL; i++)
	{
		g_print("%d, engine_name = [%s]\n", i, this_ptim->engine_list[i]);
		str = g_key_file_get_string (keyfile, "IM Group", this_ptim->engine_list[i], NULL);

		if(strcmp(str, "NONE") != 0)
		{
			this_ptim->helper_list[j] = g_strdup(str);
			j++;
		}
	}

	for (i = 0; this_ptim->helper_list[i] != NULL; i++)
	{
		g_print("%d, helper_name = [%s]\n", i, this_ptim->helper_list[i]);
	}

	this_ptim->ptim_helpers_count = i;
	g_print("this_ptim->ptim_engine_count: %d\n", this_ptim->ptim_engines_count);
	g_print("this_ptim->ptim_helper_count: %d\n", this_ptim->ptim_helpers_count);

	this_ptim->engine_module = im_ptim_get_engine(this_ptim->engine_list[0]);
	this_ptim->helper_module = im_ptim_get_helper(this_ptim->helper_list[0]);

	/* register valid IM Engine and Helper */
	engine_register(&gtk_imm_functions, &this_ptim->engine_module->engine_func);
	helper_register(&gtk_imm_functions, &this_ptim->helper_module->helper_func);

	/* Input Method Engine initialization */
	this_ptim->engine_module->engine_func.initialize();

	this_ptim->ptim_engines_array = g_array_sized_new(TRUE, TRUE, sizeof(IME_InputMethod), this_ptim->ptim_engines_count);
	for (i = 0; i < this_ptim->ptim_engines_count; i++)
	{
		im_info = (IME_InputMethod *) g_malloc(sizeof(IME_InputMethod));
		for(j = 0; j < this_ptim->engine_module->engine_func.get_number_of_input_methods(); j++)
		{
			if (this_ptim->engine_module->engine_func.get_input_method_info(j, im_info) == IME_TRUE)
			{
				if(strcmp(this_ptim->engine_list[i], im_info->name) == 0)
				{
					g_array_append_val(this_ptim->ptim_engines_array, *im_info);
					break;
				}
				else continue;
			}
		}
		g_free(im_info);
	}

	this_ptim->ptim_helpers_array = g_array_sized_new(TRUE, TRUE, sizeof(IME_Helper_Info), this_ptim->ptim_helpers_count);
	for(i = 0; i < this_ptim->ptim_helpers_count; i++)
	{
		imh_info = (IME_Helper_Info *) g_malloc(sizeof(IME_Helper_Info));
		for(j = 0; j < this_ptim->helper_module->helper_func.get_number_of_helper(); j++)
		{
			if (this_ptim->helper_module->helper_func.get_helper_info(i, imh_info) == 1)
			{
				if(strcmp(this_ptim->helper_list[i], imh_info->name) == 0)
				{
					g_array_append_val(this_ptim->ptim_helpers_array, *imh_info);
					break;
				}
				else continue;
			}
		}
		g_free(imh_info);
	}


	/* load key to switch input method from configuration file */
	str = g_key_file_get_string (keyfile, "Configuration", "SwitchKey", NULL);
	if (str)
	{
		g_strstrip (str);
		i = gdk_keyval_from_name (str);
		if (i != GDK_VoidSymbol)
			im_switch_key = i;
		g_free (str);
	}
	
	/*select default IM Engine : number input */
	this_ptim->ptim_engines_index = 0;	
	gtk_im_context_ptim_switch_input_method(this);

	return;
}

static void 
gtk_im_context_ptim_finalize(GObject * this)
{
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);
	GtkIMContext * super_class;
			
	g_source_remove(longkeypresstimer);
	g_key_file_free(keyfile);
	
	if (TRUE)
	{
		iac_context_t * context = im_ptim_get_iac_contact();
		iac_arguments_t * args = NULL;
				
		args = iac_arguments_new();
		
		iac_arguments_append_args(args,
															IAC_TYPE_INT32,
															0,
															IAC_TYPE_INVALID);	
		
		iac_emit_signal(context,
										PTIM_DBUS_OBJECT,
										PTIM_DBUS_INTERFACE,
										PTIM_DBUS_METHOD,
										args);
										
		iac_arguments_free(args);
	}

	this_ptim->helper_module->helper_func.finalize();
	this_ptim->engine_module->engine_func.finalize();

	/* Super class finalizatoin. */
	super_class = g_type_class_peek(g_type_parent(GTK_TYPE_IM_CONTEXT_PTIM));
	G_OBJECT_CLASS(super_class)->finalize(this);

	return;
}

static void 
gtk_im_context_ptim_set_client_window(GtkIMContext * this, GdkWindow *client_window)
{
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);
	
	if (this_ptim->client_window != NULL)
	{
		g_object_unref(this_ptim->client_window);
		this_ptim->client_window = NULL;
	}

	if (client_window != NULL) 
	{
		this_ptim->client_window = g_object_ref(client_window);
	}

	return;
}

static gboolean 
gtk_im_context_ptim_long_key_press(GtkIMContext * this)
{
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);

	char str[2];
	str[1] = '\0';

	switch(lastkeyval)
	{
		case GDK_0:		
		case GDK_1:		
		case GDK_2:			
		case GDK_3:			
		case GDK_4:		
		case GDK_5:			
		case GDK_6:			
		case GDK_7:		
		case GDK_8:			
		case GDK_9: str[0] = lastkeyval - GDK_0 + '0'; break;
			
		case GDK_comma:	str[0] = '*'; break;
		case GDK_period: str[0] = '#'; break;
		case GDK_BackSpace: 											/*clean all data in an entry*/

		default: longkeyhandled = TRUE; return FALSE;
	}

	this_ptim->engine_module->engine_func.direct_commit(this_ptim->input_context,
							 (void *)str); 

	longkeyhandled = TRUE;
	return FALSE;
}

static gboolean 
gtk_im_context_ptim_filter_keypress(GtkIMContext * this, GdkEventKey *event)
{
	g_print("filter_keypress event->keyval = %d\n", event->keyval);
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);
	gboolean handled = TRUE;
	IME_KeyEvent keyEvent;
	int index = 0;
	lastkeyval = event->keyval;
	global_ic_ptim = this_ptim;

	if(gtk_im_context_ptim_is_fcitx_input(this) && (event->keyval == GDK_bracketleft || event->keyval == GDK_bracketright || (event->keyval >= 48 && event->keyval <= 57) || (event->keyval >= 97 && event->keyval <= 122)))
	{
		keyEvent.code		= event->keyval;
		keyEvent.unicode	= event->keyval;
		keyEvent.mask		= 0;

		if(event->keyval >= 48 && event->keyval <= 57)
		{
			this_ptim->engine_module->engine_func.select_candidate(this_ptim->input_context, event->keyval);
		}
		else if(event->keyval == GDK_bracketleft)
		{
			this_ptim->engine_module->engine_func.previous_candidates_page(this_ptim->input_context);
		}
		else if(event->keyval == GDK_bracketright)
		{
			this_ptim->engine_module->engine_func.next_candidates_page(this_ptim->input_context);
		}
		else
		{
			this_ptim->engine_module->engine_func.add_element(this_ptim->input_context, (void *)&keyEvent);
		}
		handled = TRUE;
		return TRUE;
	}


	if(event->type == GDK_KEY_PRESS)
	{	
		longkeyhandled = FALSE;
		longkeypresstimer = g_timeout_add(LONG_KEY_PRESS_INTERVAL, gtk_im_context_ptim_long_key_press, this);
		if(event->keyval != GDK_BackSpace && event->keyval != GDK_Up && event->keyval != GDK_Down &&
				event->keyval != GDK_Left && event->keyval != GDK_Right && event->keyval != GDK_Return) 
		{
			return TRUE;
		}
	} 
	else if (event->type == GDK_KEY_RELEASE) 
	{
		if(!longkeyhandled) g_source_remove(longkeypresstimer);		/*destroy timer*/
		else return TRUE;			/*long key press has been handled in gtk_im_context_ptim_long_key_press()*/

		if(event->keyval == GDK_BackSpace || event->keyval == GDK_Up || event->keyval == GDK_Down || 
				event->keyval == GDK_Left || event->keyval == GDK_Right || event->keyval == GDK_Return)
		{
		 return TRUE;
		}
	}


	/* input method switch triggered? */
	if (event->keyval == im_switch_key)
	{
		this_ptim->ptim_engines_index = (this_ptim->ptim_engines_index + 1) % this_ptim->ptim_engines_count;
		gtk_im_context_ptim_switch_input_method(this);
		handled = TRUE;
	}
	else /* not the switch key, handle like specified... */
	{
		switch(event->keyval)
		{
			case GDK_comma:				/*input method switching * key*/
				g_print("input key: GDK_comma\n");
				handled = FALSE;
				break;
		
			case GDK_Left:		
				g_print("input key: GDK_Left\n");

				if (gtk_im_context_ptim_is_num_input(this))
				{
					handled = FALSE;
					break;
				}

				if (this_ptim->status == IME_PREEDIT && this_ptim->preedit_candidates_list != NULL)
				{
					/* no input element, use the original function the "Left" key */
					if (this_ptim->preedit_candidates_list->nr_candidates == 0)
					{
						handled = FALSE;
						break;
					}

					if (this_ptim->preedit_candidates_list->focused_candidate > 0)
					{
						this_ptim->preedit_candidates_list->focused_candidate--;
					}
					else
					{
						/* if move to the first candidate of this page, rewind */
						this_ptim->preedit_candidates_list->focused_candidate = this_ptim->preedit_candidates_list->nr_candidates - 1;
					}
					this_ptim->engine_module->engine_func.select_preedit_candidate(this_ptim->input_context, 
																										this_ptim->preedit_candidates_list->focused_candidate);
				}

				else if (this_ptim->candidates_list != NULL)
				{
					if (this_ptim->status == IME_ASSOCIATE)
					{
						this_ptim->numeric_select = TRUE;
						this_ptim->helper_module->helper_func.notify_numeric_display(this_ptim->numeric_select);
					}
					
					if (this_ptim->candidates_list->focused_candidate > 0)
					{
						this_ptim->candidates_list->focused_candidate--;
					}
					else
					{
						this_ptim->candidates_list->focused_candidate = this_ptim->candidates_list->nr_candidates - 1;
					}
					this_ptim->helper_module->helper_func.show_candidates_list(this_ptim->candidates_list);
				}

				else
				{
					handled = FALSE;
					break;
				}
				
				break;
				
			case GDK_Right:
				g_print("input key: GDK_Right\n");

				if (gtk_im_context_ptim_is_num_input(this))
				{
					handled = FALSE;
					break;
				}

				if (this_ptim->status == IME_PREEDIT && this_ptim->preedit_candidates_list != NULL)
				{
					/* no input element, use the original function the "Right" key */
					if (this_ptim->preedit_candidates_list->nr_candidates == 0)
					{
						handled = FALSE;
						break;
					}
					
					if (this_ptim->preedit_candidates_list->focused_candidate < this_ptim->preedit_candidates_list->nr_candidates - 1)
					{
						this_ptim->preedit_candidates_list->focused_candidate++;
					}
					else
					{
						/* if move to the first candidate of this page, rewind */
						this_ptim->preedit_candidates_list->focused_candidate = 0;
					}
					this_ptim->engine_module->engine_func.select_preedit_candidate(this_ptim->input_context, 
																											this_ptim->preedit_candidates_list->focused_candidate);
				}

				else if (this_ptim->candidates_list != NULL)
				{
					if (this_ptim->status == IME_ASSOCIATE)
					{
						this_ptim->numeric_select = TRUE;
						this_ptim->helper_module->helper_func.notify_numeric_display(this_ptim->numeric_select);
					}
					
					if (this_ptim->candidates_list->focused_candidate < this_ptim->candidates_list->nr_candidates - 1)
					{
						this_ptim->candidates_list->focused_candidate++;
					}
					else
					{
						/* if move to the last candidate of this page, rewind */
						this_ptim->candidates_list->focused_candidate = 0;
					}
					this_ptim->helper_module->helper_func.show_candidates_list(this_ptim->candidates_list);
				}
				else
				{
					handled = FALSE;
					break;
				}
				break;

			case GDK_Up:
				g_print("input key: GDK_Up\n");

				if (gtk_im_context_ptim_is_num_input(this))
				{
					handled = FALSE;
					break;
				}
				if (this_ptim->status == IME_PREEDIT)
				{
					if(this_ptim->preedit_candidates_list == NULL)
					{
						handled = FALSE;
						break;
					}
					/* no input element, use the original function the "Up" key */
					if (this_ptim->preedit_candidates_list->nr_candidates == 0)
					{
						handled = FALSE;
						break;
					}
					
					this_ptim->engine_module->engine_func.previous_preedit_candidates_page(this_ptim->input_context);
				}
				else 
	/*			if (this_ptim->status == IME_NORMAL || this_ptim->status == IME_ASSOCIATE)
			*/{				
					if( this_ptim->candidates_list == NULL || this_ptim->candidates_list->nr_candidates == 0)
					{
						handled = FALSE;
						break;
					}
					this_ptim->engine_module->engine_func.previous_candidates_page(this_ptim->input_context);
					this_ptim->candidates_list->focused_candidate = 0;
				}
				break;
				
			case GDK_Down:
				g_print("input key: GDK_Down\n");

				if(gtk_im_context_ptim_is_num_input(this)) 
				{
					handled = FALSE;
					break;
				}
				
				if (this_ptim->status == IME_PREEDIT)
				{
					if(this_ptim->preedit_candidates_list == NULL)
					{
						handled = FALSE;
						break;
					}
					/* no input element, use the original function the "Down" key */
					if (this_ptim->preedit_candidates_list->nr_candidates == 0)
					{
						handled = FALSE;
						break;
					}
				
					this_ptim->engine_module->engine_func.next_preedit_candidates_page(this_ptim->input_context);
					this_ptim->preedit_candidates_list->focused_candidate = 0;
				}
				else 
	/*			if (this_ptim->status == IME_NORMAL || this_ptim->status == IME_ASSOCIATE)
		*/	{				
					if( this_ptim->candidates_list == NULL || this_ptim->candidates_list->nr_candidates == 0)
					{
						handled = FALSE;
						break;
					}
					
					this_ptim->engine_module->engine_func.next_candidates_page(this_ptim->input_context);
					this_ptim->candidates_list->focused_candidate = 0;
				}
				break;
				
			case GDK_0:
			/*In Pinyin or Latin input method, we accept GDK_0 as a space character*/
			
				if(!gtk_im_context_ptim_is_num_input(this))		/*key 0 means ' ' when current IM is not num*/
				{
					char str[2];
					str[1] = '\0';
					/*take num input */
					str[0] = ' ';
					this_ptim->engine_module->engine_func.direct_commit(this_ptim->input_context,
								 (void *)str); 

					handled = TRUE;
					break;
				}
				else 
				{
					char str[2];
					str[1] = '\0';
					/*take num input */
					str[0] = '0';
					this_ptim->engine_module->engine_func.direct_commit(this_ptim->input_context,
								 (void *)str); 

					handled = TRUE;
					break;
				}

			case GDK_1:	
				
			case GDK_2:
				
			case GDK_3:
				
			case GDK_4:
				
			case GDK_5:
				
			case GDK_6:
				
			case GDK_7:
				
			case GDK_8:
				
			case GDK_9:
				g_print("input key: GDK_%d\n", event->keyval - GDK_0);
			
				keyEvent.code		= event->keyval;
				keyEvent.unicode	= event->keyval;
				keyEvent.mask		= 0;

				if (gtk_im_context_ptim_is_num_input(this))
				{
					char str[2];
					str[1] = '\0';
					/*take num input */
					str[0] = event->keyval - GDK_0 + '0';
					this_ptim->engine_module->engine_func.direct_commit(this_ptim->input_context,
								 (void *)str); 
					handled = TRUE;
					break;
				}
				
				if(gtk_im_context_ptim_is_TS_eng_input(this))
				{
					handled = TRUE;
					break;
				}
			
				/*Two-step input method*/
				if(this_ptim->status == IME_PREEDIT)
				{
					g_print("status: IME_PREEDIT\n");
					
					this_ptim->numeric_select = FALSE;
					this_ptim->helper_module->helper_func.notify_numeric_display(this_ptim->numeric_select);
					if (this_ptim->engine_module->engine_func.add_element(this_ptim->input_context, (void *)&keyEvent) == IME_FALSE)
					{
						fprintf(stdout, "add_element fail\n"); 
						handled = FALSE;
					}
					else
					{
						fprintf(stdout, "add_element succeed\n"); 
						if(event->keyval == GDK_1)
						{/*accept it as a punctuation, enable the numeric selection*/
							this_ptim->numeric_select = TRUE;
							this_ptim->helper_module->helper_func.notify_numeric_display(this_ptim->numeric_select);
							
							this_ptim->status = IME_PUNCT;
							this_ptim->helper_module->helper_func.notify_ime_status(this_ptim->status);
							this_ptim->helper_module->helper_func.show_candidates_list(this_ptim->candidates_list);						
						}
						else
						{
							this_ptim->numeric_select = FALSE;
							this_ptim->helper_module->helper_func.notify_numeric_display(this_ptim->numeric_select);
						}
					}
				}
				else if(this_ptim->status == IME_NORMAL || this_ptim->status == IME_PUNCT)
				{
					index = event->keyval - GDK_1;
					
					if (index >= 0 && index <= this_ptim->candidates_list->nr_candidates - 1)
					{
						this_ptim->numeric_select = FALSE;
						this_ptim->helper_module->helper_func.notify_numeric_display(this_ptim->numeric_select);

						this_ptim->status = IME_ASSOCIATE;
						this_ptim->engine_module->engine_func.select_candidate(this_ptim->input_context, index);					
						
						this_ptim->helper_module->helper_func.notify_ime_status(this_ptim->status);
					}
					else
					{
						/*invalid choice, nothing to do currently,
						 * strongly recommented that we return to pre-edit mode
						 *  */
						g_print("invalid choice\n");
					}		
				}
				else if(this_ptim->status == IME_ASSOCIATE)
				{
					g_print("status: IME_ASSOCIATE\n");
					index = event->keyval - GDK_1;
					
					if(this_ptim->numeric_select == TRUE)
					{
						if (index >= 0 && index <= this_ptim->candidates_list->nr_candidates - 1)
						{
							this_ptim->numeric_select = FALSE;
							g_print("%s(): to reset numeric_select\n", __FUNCTION__);
							this_ptim->helper_module->helper_func.notify_numeric_display(this_ptim->numeric_select);

							this_ptim->engine_module->engine_func.select_candidate(this_ptim->input_context, index);					
							this_ptim->status = IME_ASSOCIATE;
							this_ptim->helper_module->helper_func.notify_ime_status(this_ptim->status);
						}
						else
						/*invalid choice, nothing to do currently,
						 * strongly recommented that we return to pre-edit mode
						 * */
						{
							g_print("invalid choice\n");
						}
					}
					else
					{
						this_ptim->status = IME_PREEDIT;
						this_ptim->helper_module->helper_func.notify_ime_status(this_ptim->status);
						this_ptim->numeric_select = FALSE;
						this_ptim->helper_module->helper_func.notify_numeric_display(this_ptim->numeric_select);
						if (this_ptim->engine_module->engine_func.add_element(this_ptim->input_context, (void *)&keyEvent) == IME_FALSE)
						{
							fprintf(stdout, "add_element fail\n"); 
							handled = FALSE;
						}
						else
						{
							fprintf(stdout, "add_element succeed\n");
							if( event->keyval == GDK_1 )
							{/*accept it as a punctuation, enable the numeric selection*/
								this_ptim->numeric_select = TRUE;
								this_ptim->helper_module->helper_func.notify_numeric_display(this_ptim->numeric_select);
								
								this_ptim->status = IME_PUNCT;
								this_ptim->helper_module->helper_func.notify_ime_status(this_ptim->status);
								this_ptim->helper_module->helper_func.show_candidates_list(this_ptim->candidates_list);						
							}						 
						}
					}								
				}
				break;
				
			case GDK_Return:				/*OK key*/
				g_print("input key: GDK_Return\n");

			  if (gtk_im_context_ptim_is_num_input(this) || gtk_im_context_ptim_is_TS_eng_input(this))
				{
					handled = FALSE;
					break;
				}

				keyEvent.code		= event->keyval;
				keyEvent.unicode	= event->keyval;
				keyEvent.mask		= 0;

				if ((this_ptim->status != IME_PUNCT) && 
							(this_ptim->preedit_candidates_list == NULL || this_ptim->candidates_list == NULL))
				{		
	/*			handled = FALSE;*/
					break;
				}
				break;

			case GDK_period:									/*symbol table*/
				g_print("input key: GDK_period\n");
			/*Mapping to '#' in the mobile keyboard*/ 

				break;
				
			case GDK_BackSpace:
				g_print("input key: GDK_BackSpace\n");
				
				keyEvent.code = event->keyval;
				keyEvent.unicode = event->keyval;
				keyEvent.mask = 0;
				if (gtk_im_context_ptim_is_num_input(this))
				{
					/*take num input*/
					handled = FALSE;
					break;
				}

				else 
				{
					handled = FALSE;
					break;
				}						
				break;

			default:
				g_print("input key: not known\n");
			
				handled = FALSE;
				break;
		}
	 }
	
	return handled;	
}

static void 
gtk_im_context_ptim_focus_in(GtkIMContext *this)
{
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);

	global_ic_ptim = GTK_IM_CONTEXT_PTIM(this); 
	
	g_print("%s(): entering, 0x%08x\n",
					__FUNCTION__,
					(guint32)this);

	this_ptim->preedit_candidates_list	= NULL;		/*blank context*/
	this_ptim->candidates_list	= NULL;

	return;
}

static void 
gtk_im_context_ptim_focus_out(GtkIMContext *this)
{
	GtkIMContextPTIM *	this_ptim	= GTK_IM_CONTEXT_PTIM(this);
	
	global_ic_ptim = GTK_IM_CONTEXT_PTIM(this); 
	
	g_print("%s(): entering, 0x%08x\n",
					__FUNCTION__,
					(guint32)this);
	
	this_ptim->ptim_engines_index = 0;	
	gtk_im_context_ptim_switch_input_method(this);

	return;
}

static void 
gtk_im_context_ptim_set_cursor_location(GtkIMContext * this, GdkRectangle * area)
{
/*
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);
	gint x, y;
	gint cl_width, cl_height;

//	g_print("%s(): entering\n", __FUNCTION__);
	if (this_ptim->client_window != NULL) 
	{
		gdk_window_get_origin(this_ptim->client_window, &x, &y);
		gdk_window_get_size(this_ptim->client_window, &cl_width, &cl_height);

		this_ptim->helper_module->helper_func.notify_client_size(x, y, cl_width, cl_height);
	}
*/
	return;
}

static void 
gtk_im_context_ptim_reset(GtkIMContext *this)
{
	GtkIMContextPTIM * this_ptim = GTK_IM_CONTEXT_PTIM(this);

	global_ic_ptim = GTK_IM_CONTEXT_PTIM(this); 
	
	return;
}

static void 
gtk_im_context_ptim_get_preedit_string(GtkIMContext *this, gchar **str, PangoAttrList **attrs, gint *cursor_pos)
{
	global_ic_ptim = GTK_IM_CONTEXT_PTIM(this); 

	if (str)
	{
		GString * s;
		s = g_string_new("");
		*str = g_string_free(s, FALSE);
	}
	if (attrs) 
	{
		*attrs = pango_attr_list_new();
	}	
	if (cursor_pos)
	{
		*cursor_pos = 0;
	}

	return;
}

static void 
gtk_im_context_ptim_class_init(gpointer class, gpointer class_data)
{
	/* GtkIMContextPTIMClass specific methods initialization. */
	
	/* Override super class methods. */
	GTK_IM_CONTEXT_CLASS(class)->set_client_window			= gtk_im_context_ptim_set_client_window;
	GTK_IM_CONTEXT_CLASS(class)->filter_keypress				= gtk_im_context_ptim_filter_keypress;
	GTK_IM_CONTEXT_CLASS(class)->reset										= gtk_im_context_ptim_reset;
	GTK_IM_CONTEXT_CLASS(class)->get_preedit_string		= gtk_im_context_ptim_get_preedit_string;
	GTK_IM_CONTEXT_CLASS(class)->focus_in								= gtk_im_context_ptim_focus_in;
	GTK_IM_CONTEXT_CLASS(class)->focus_out								= gtk_im_context_ptim_focus_out;
	GTK_IM_CONTEXT_CLASS(class)->set_cursor_location	= gtk_im_context_ptim_set_cursor_location;

	G_OBJECT_CLASS(class)->finalize												= gtk_im_context_ptim_finalize;
	
	return;
}

/*
 * GType and GTypeModule Registeration
 */
GType im_context_ptim_type = 0;

void
gtk_im_context_ptim_register_type(GTypeModule * type_module)
{	
	if (!im_context_ptim_type)
	{
		static const GTypeInfo gtk_im_context_ptim_info = 
		{
			sizeof(GtkIMContextPTIMClass),
			NULL,                       /* base_init */
			NULL,                       /* base_finalize */
			gtk_im_context_ptim_class_init,
			NULL,                       /* class_finalize */
			NULL,                       /* class_data */
			sizeof(GtkIMContextPTIM),
			0,                          /* n_preallocs */
			gtk_im_context_ptim_init
		};

		im_context_ptim_type = g_type_module_register_type(
															type_module,
															GTK_TYPE_IM_CONTEXT,
															"GtkIMContextPTIM",
															&gtk_im_context_ptim_info,
															0);
	}
	return;
}

GtkIMContext *
gtk_im_context_ptim_new (void)
{
  GtkIMContextPTIM *result;

  result = g_object_new(GTK_TYPE_IM_CONTEXT_PTIM, NULL);
	
  return GTK_IM_CONTEXT(result);
}
