/*	immanager - Text Input Method Manager
 *
 *  imptim.c - 
 *
 *	Authors: YE Nan <nan.ye@orange-ftgourp.com>,
 *           LUO Lei <Srella@163.com>
 *
 *	
 *	This software and associated documentation files (the "Software") 
 *	are copyright (C) 2005 LiPS Linux Phone Standards Forum [FranceTelecom] 
 *	All Rights Reserved. 
 *
 *	A copyright license is hereby granted for redistribution and use of 
 *	the Software in source and binary forms, with or without modification, 
 *	provided that the following conditions are met: 
 *	- Redistributions of source code must retain the above copyright notice, 
 *	this copyright license and the following disclaimer. 
 *  - Redistributions in binary form must reproduce the above copyright 
 * 	notice, this copyright license and the following disclaimer in the 
 *	documentation and/or other materials provided with the distribution. 
 *	- Neither the name of LiPS  nor the names of its Members may be used 
 *	to endorse or promote products derived from the Software without 
 *	specific prior written permission. 
 *
 *	A patent license for any Necessary Claims owned by Members of LiPS Forum 
 *	to make, have made, use, import, offer to sell, lease and sell or otherwise 
 *	distribute any implementation compliant with the any specification adopted 
 *	by the LiPS Forumcan be obtained from the respective Members on reasonable 
 *	and non-discriminatory terms and conditions and under reciprocity, as 
 *	regulated in more detail in the Internal Policy of the LiPS Forum. 
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER, ITS MEMBERS AND CONTRIBUTORS 
 *	"AS IS", AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *	THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE 
 *	AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER, 
 *	ITS MEMBERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 *	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 *	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *	POSSIBILITY OF SUCH DAMAGE. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <iac.h>

#include <gtk/gtkimmodule.h>

#include "gtkimcontextptim.h"
#include "ptimhelpermodule.h"
#include "ptimenginemodule.h"

extern IMM_Functions	gtk_imm_functions;

static GPtrArray * engine_modules	= NULL;
static GPtrArray * helper_modules	= NULL;

static GHashTable * engine_maps		= NULL;
static GHashTable * helper_maps		= NULL;

gchar ** engine_names = NULL;
gchar ** helper_names = NULL;

iac_context_t * context = NULL;

static const GtkIMContextInfo ptim_info = {
  "ptim",
  "Phone Text Input Method",
  "gtk+",
	"",
  "en:zh"
};

static const GtkIMContextInfo *info_list[] = {
  &ptim_info
};

static void
ptim_engine_info_print (IME_InputMethod * info)
{
	g_print("Engine:\n");
	g_print("|- uuid        : %s\n", (info->uuid ? info->uuid : "(null)"));
	g_print("|- encoding    : %s\n", (info->encoding ? info->encoding : "(null)"));
	g_print("|- name        : %s\n", (info->name ? info->name : "(null)"));
	g_print("|- icon        : %s\n", (info->icon ? info->icon : "(null)"));
	g_print("|- languages   : %s\n", (info->languages ? info->languages : "(null)"));
	g_print("|- help        : %s\n", (info->help ? info->help : "(null)"));
	g_print("|- copyright   : %s\n", (info->copyright ? info->copyright : "(null)"));
	g_print("|- version     : %s\n", (info->version ? info->version : "(null)"));
	g_print("|- capability  : 0x%08x\n", info->capability_mask);

	return;
}

static void
ptim_helper_info_print (IME_Helper_Info * info)
{
	g_print("Helper:\n");
	g_print("|- uuid        : %s\n", (info->uuid ? info->uuid : "(null)"));
	g_print("|- encoding    : %s\n", (info->encoding ? info->encoding : "(null)"));
	g_print("|- name        : %s\n", (info->name ? info->name : "(null)"));
	g_print("|- description : %s\n", (info->description ? info->description : "(null)"));
	g_print("|- icon        : %s\n", (info->icon ? info->icon : "(null)"));
	g_print("|- capability  : 0x%08x\n", info->capability_mask);

	return;
}

static void 
ptim_module_map_print (gpointer key, gpointer value, gpointer user_data)
{
	g_print("%s(): name = %s, module = [0x%08x]\n",
					__FUNCTION__,
					(gchar *)key,
					(guint32)value);
					
	return;
}

static void
ptim_engine_module_free (PtimHelperModule * engine_module)
{
	return;
}

static void
ptim_helper_module_free (PtimHelperModule * helper_module)
{
	return;
}

static gboolean
ptim_add_engine_module (const gchar * module_path)
{
	PtimEngineModule * engine_module = NULL;

	engine_module = g_object_new(PTIM_TYPE_ENGINE_MODULE, NULL);
	engine_module->path = g_strdup(module_path);
	g_type_module_set_name(G_TYPE_MODULE(engine_module), engine_module->path);

	if (g_type_module_use(G_TYPE_MODULE(engine_module)))
	{
		if (engine_module->init)
		{
			engine_module->init(G_TYPE_MODULE(engine_module));
		}
	}
	else
	{
		return FALSE;
	}

	/*
	 * register a module, and map all the engines it supplies
	 * with their names
	 */
	g_ptr_array_add(engine_modules, (gpointer)engine_module);
	if (TRUE)
	{
		gint i;
		gint count;
		
		count = engine_module->engine_func.get_number_of_input_methods();
		for (i = 0; i < count; i++)
		{
			IME_InputMethod im;
			
			engine_module->engine_func.get_input_method_info(i, &im);

			g_hash_table_replace(engine_maps,
													 g_strdup(im.name),
													 (gpointer)engine_module);
			
		}
	}	
	return TRUE;
}

static gboolean
ptim_add_helper_module (const gchar * module_path)
{
	PtimHelperModule * helper_module = NULL;

	helper_module = g_object_new(PTIM_TYPE_HELPER_MODULE, NULL);
	helper_module->path = g_strdup(module_path);
	g_type_module_set_name(G_TYPE_MODULE(helper_module), helper_module->path);
	
	if (g_type_module_use(G_TYPE_MODULE(helper_module)))
	{
		if (helper_module->init)
		{
			helper_module->init(G_TYPE_MODULE(helper_module));
		}
	}
	else
	{
		return FALSE;
	}

	/*register and map a new helper with its name*/
	g_ptr_array_add(helper_modules, (gpointer)helper_module);

	if (TRUE)
	{
		gint i;
		gint count;
		
		count = helper_module->helper_func.get_number_of_helper();

		for (i = 0; i < count; i++)
		{
			IME_Helper_Info info;
			
			helper_module->helper_func.get_helper_info(i, &info);

			g_hash_table_replace(helper_maps,
											 g_strdup(info.name),
											 (gpointer)helper_module);		
		}
	}			
	return TRUE;
}

/**
 * find Helper modules in specific path.
 * path is implied by PTIM_ENGINE_PATH 
 * environment variable
 */
static gboolean
ptim_find_engine_modules (const gchar * dir_path)
{
	GDir * dir = NULL;
	GError * error = NULL;
	const gchar * fname = NULL;
	
	dir = g_dir_open(dir_path, 0, &error);
	if (dir == NULL)
	{
		g_print("%s(): directory open failure, %s\n", 
						__FUNCTION__,
						error->message);
						
		g_error_free(error);
		return FALSE;
	}

	g_dir_rewind(dir);
	
	while ((fname = g_dir_read_name(dir)) != NULL)
	{
		gchar * abosulte_path = NULL;
		
		abosulte_path = g_build_filename(dir_path, fname, NULL);
		g_print("%s(): module_fname = %s\n", __FUNCTION__, abosulte_path);
		ptim_add_engine_module(abosulte_path);
		
		g_free(abosulte_path);
	}
	
	g_dir_close(dir);
	
	return TRUE;
}

/**
 * find Helper modules in specific path.
 * path is implied by PTIM_HELPER_PATH 
 * environment variable
 */
static gboolean
ptim_find_helper_modules (const gchar * dir_path)
{
	GDir * dir = NULL;
	GError * error = NULL;
	const gchar * fname = NULL;
	
	dir = g_dir_open(dir_path, 0, &error);
	if (dir == NULL)
	{
		g_print("%s(): directory open failure, %s\n", 
						__FUNCTION__,
						error->message);
						
		g_error_free(error);
		return FALSE;
	}

	g_dir_rewind(dir);
	
	while ((fname = g_dir_read_name(dir)) != NULL)
	{
		gchar * abosulte_path = NULL;
		
		abosulte_path = g_build_filename(dir_path, fname, NULL);
		g_print("%s(): module_fname = %s\n", __FUNCTION__, abosulte_path);
		ptim_add_helper_module(abosulte_path);

		g_free(abosulte_path);
	}
	
	g_dir_close(dir);
	
	return TRUE;
}

static void 
im_ptim_get_module_name (gpointer key, gpointer value, GPtrArray * names)
{
	if (key)
	{
		g_ptr_array_add(names, (gpointer)g_strdup((gchar *)key));
	}	
	return;
}

void 
im_module_init(GTypeModule * type_module)
{
	g_print("%s(): type_module = 0x%08x\n", __FUNCTION__, (guint32)type_module);

	/*
	 * register GtkIMContextPTIM,
	 * PtimEngineModule and PtimHelperModule types
	 */
	gtk_im_context_ptim_register_type(type_module);
	ptim_engine_module_register_type(type_module);
	ptim_helper_module_register_type(type_module);

	engine_modules = g_ptr_array_new();
	helper_modules = g_ptr_array_new();

	engine_maps = g_hash_table_new_full(g_str_hash,
																			g_str_equal,
																			(GDestroyNotify)g_free,
																			(GDestroyNotify)ptim_engine_module_free);
																				 
	helper_maps = g_hash_table_new_full(g_str_hash,
																			g_str_equal,
																			(GDestroyNotify)g_free,
																			(GDestroyNotify)ptim_helper_module_free);

	if (TRUE)
	{
		char * engine_module_dir = NULL;
		char * helper_module_dir = NULL;

		if ((engine_module_dir = getenv("PTIM_ENGINE_PATH")) == NULL)
		{
			g_error("%s(): set PTIM_ENGINE_PATH firstly\n", __FUNCTION__);
		}
			
		if ((helper_module_dir = getenv("PTIM_HELPER_PATH")) == NULL)
		{
			g_error("%s(): set PTIM_HELPER_PATH firstly\n", __FUNCTION__);
		}
		
		/*
		 * loading engines and helpers
		 * from modules path
		 */ 

		ptim_find_engine_modules(engine_module_dir);
		ptim_find_helper_modules(helper_module_dir);
	}
	
	/*
	 * make up a engine name list
	 */
	if (TRUE)
	{
		GPtrArray * names = g_ptr_array_new();
		
		g_hash_table_foreach(engine_maps,
												 (GHFunc)im_ptim_get_module_name,
												 (gpointer)names);
												 
		g_ptr_array_add(names, (gpointer)NULL);
		
		engine_names = (gchar **) g_ptr_array_free(names, FALSE);
	}

	if (TRUE)
	{
		GPtrArray * names = g_ptr_array_new();
		
		g_hash_table_foreach(helper_maps,
												 (GHFunc)im_ptim_get_module_name,
												 (gpointer)names);
									 
		g_ptr_array_add(names, (gpointer)NULL);
		
		helper_names = (gchar **) g_ptr_array_free(names, FALSE);
	}

	context = iac_initialize();

	return;
}

void 
im_module_exit (void)
{
	gint i;
	/*
	 * unuse all engine modules
	 */
	for (i = 0; i < engine_modules->len; i++)
	{
		PtimEngineModule * module = PTIM_ENGINE_MODULE(g_ptr_array_index(engine_modules, i));
			
		if (module->exit)
		{
			module->exit(G_TYPE_MODULE(module));
		}
		
		g_type_module_unuse(G_TYPE_MODULE(module));
	}
	
	/*
	 * unuse all helper modules
	 */
	for (i = 0; i < helper_modules->len; i++)
	{
		PtimEngineModule * module = PTIM_ENGINE_MODULE(g_ptr_array_index(helper_modules, i));
			
		if (module->exit)
		{
			module->exit(G_TYPE_MODULE(module));
		}
		
		g_type_module_unuse(G_TYPE_MODULE(module));
	}
	
	g_strfreev(engine_names);
	g_strfreev(helper_names);
	
	g_ptr_array_free(engine_modules, TRUE);
	g_ptr_array_free(helper_modules, TRUE);

	g_hash_table_destroy(engine_maps);
	g_hash_table_destroy(helper_maps);
	
	iac_deinitialize(context);
	
	return;
}

void 
im_module_list (const GtkIMContextInfo *** contexts, int * n_contexts)
{
	*contexts = info_list;
	*n_contexts = G_N_ELEMENTS(info_list); 
	return;
}

GtkIMContext * 
im_module_create (const gchar * context_id)
{
	g_print("\n%s(): entering   context_id = %s\n", __FUNCTION__, context_id);

  if (g_strcasecmp(context_id, "ptim") == 0)
  {
		return gtk_im_context_ptim_new();
  }
	return NULL;
}

G_CONST_RETURN gchar **
im_ptim_get_engine_list (void)
{
	return (const gchar **) engine_names;
}

G_CONST_RETURN gchar **
im_ptim_get_helper_list (void)
{
	return (const gchar **) helper_names;
}

PtimEngineModule *
im_ptim_get_engine (const gchar * engine_name)
{
	return g_hash_table_lookup(engine_maps, engine_name);
}

PtimHelperModule *
im_ptim_get_helper (const gchar * helper_name)
{
	return g_hash_table_lookup(helper_maps, helper_name);
}

iac_context_t *
im_ptim_get_iac_contact(void)
{
	return context;
}
