/*	immanager - Text Input Method Manager
 *
 *  ptimenginemodule.c - 
 *
 *	Authors: YE Nan <nan.ye@orange-ftgourp.com>,
 *           LUO Lei <Srella@163.com>
 *
 *	
 *	This software and associated documentation files (the "Software") 
 *	are copyright (C) 2005 LiPS Linux Phone Standards Forum [FranceTelecom] 
 *	All Rights Reserved. 
 *
 *	A copyright license is hereby granted for redistribution and use of 
 *	the Software in source and binary forms, with or without modification, 
 *	provided that the following conditions are met: 
 *	- Redistributions of source code must retain the above copyright notice, 
 *	this copyright license and the following disclaimer. 
 *  - Redistributions in binary form must reproduce the above copyright 
 * 	notice, this copyright license and the following disclaimer in the 
 *	documentation and/or other materials provided with the distribution. 
 *	- Neither the name of LiPS  nor the names of its Members may be used 
 *	to endorse or promote products derived from the Software without 
 *	specific prior written permission. 
 *
 *	A patent license for any Necessary Claims owned by Members of LiPS Forum 
 *	to make, have made, use, import, offer to sell, lease and sell or otherwise 
 *	distribute any implementation compliant with the any specification adopted 
 *	by the LiPS Forumcan be obtained from the respective Members on reasonable 
 *	and non-discriminatory terms and conditions and under reciprocity, as 
 *	regulated in more detail in the Internal Policy of the LiPS Forum. 
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER, ITS MEMBERS AND CONTRIBUTORS 
 *	"AS IS", AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *	THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE 
 *	AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER, 
 *	ITS MEMBERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 *	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 *	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *	POSSIBILITY OF SUCH DAMAGE. 
 */

#include <stdio.h>

#include <gtk/gtk.h>
#include <gmodule.h>

#include <standard_im.h>

#include "ptimenginemodule.h"

gboolean (*engine_register) (IMM_Functions * imm, IME_Functions * ime);

GType ptim_engine_module_type 			= 0;
static GObjectClass *parent_class	= NULL;

/*
 * Load one module from engine modules path
 */
static gboolean
ptim_engine_module_load (GTypeModule * module)
{
  PtimEngineModule * engine_module = PTIM_ENGINE_MODULE(module);
  
  engine_module->library = g_module_open (engine_module->path, 0);/////////////////////
  if (!engine_module->library)
	{
		g_print("%s(): error, %s\n", __FUNCTION__,  g_module_error());
		return FALSE;
	}
	
	/*
	 * symbol imh_register_module() to
	 * local funtion pointer
	 */  
	if (!g_module_symbol(engine_module->library, 
											 "ime_register_module",
											 (gpointer *)&engine_register))
	{
		g_print("%s(): error, %s\n", __FUNCTION__, g_module_error());
		g_module_close (engine_module->library);
      
		return FALSE;
	}
	
	/*
	 * if the module needs to register
	 * some GType, call its own initialize
	 * functions
	 */
	g_print("%s(): func ---> engine_register = 0x%0x8\n", __FUNCTION__, (guint32)(engine_register));
	engine_register(&(engine_module->manager_func), &(engine_module->engine_func));
	
	if (!g_module_symbol(engine_module->library, 
											"engine_module_init",
											(gpointer *)&engine_module->init))
	{
		engine_module->init = NULL;
	}
	
	if (!g_module_symbol(engine_module->library, 
											"engine_module_exit",
											(gpointer *)&engine_module->exit))
	{
		engine_module->exit = NULL;
	}	
	
	return TRUE;
}

static void
ptim_engine_module_unload (GTypeModule * module)
{
  PtimEngineModule * engine_module = PTIM_ENGINE_MODULE(module);
	
	g_module_close(engine_module->library);
	engine_module->library = NULL;
	
	engine_module->init = NULL;
	engine_module->exit = NULL;
  
  return;
}

/* This only will ever be called if an error occurs during
 * initialization
 */
static void
ptim_engine_module_finalize (GObject * object)
{
  PtimEngineModule * engine_module = PTIM_ENGINE_MODULE(object);

  g_free(engine_module->path);

  parent_class->finalize (object);
  
  return;
}

static void
ptim_engine_module_class_init (PtimEngineModuleClass *class)
{
  GTypeModuleClass * module_class	= G_TYPE_MODULE_CLASS(class);
  GObjectClass * gobject_class 		= G_OBJECT_CLASS(class);

  parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(class));
  
  module_class->load			= ptim_engine_module_load;
  module_class->unload		= ptim_engine_module_unload;

  gobject_class->finalize	= ptim_engine_module_finalize;
  
  return;
}

void
ptim_engine_module_register_type (GTypeModule * type_module)
{

  if (!ptim_engine_module_type)
	{
		static const GTypeInfo engine_module_info = {
  		sizeof (PtimEngineModuleClass),
 			NULL,           /* base_init */
			NULL,           /* base_finalize */
			(GClassInitFunc) ptim_engine_module_class_init,
			NULL,           /* class_finalize */
			NULL,           /* class_data */
			sizeof (PtimEngineModule),
			0,              /* n_preallocs */
			NULL,           /* instance_init */
		};

		ptim_engine_module_type =	g_type_module_register_type(type_module,
																													G_TYPE_TYPE_MODULE,
																													"PtimEngineModule",
																													&engine_module_info,
																													0);
	} 
  return;
}
